/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-linux.h"
#include "nv-vm.h"

#if defined(NV_DBG_MEM)
#define NV_DEFAULT_LIST_PAGE_COUNT 10
void
nv_vm_list_page_count(nv_pte_t *page_list, unsigned long num_pages)
{
    nv_pte_t *page_ptr = page_list;
    int i;

    if (page_ptr == NULL)
        return;

    nv_printf(NV_DBG_MEMINFO, "NVRM: VM:    page_table at 0x%p, %d pages\n", 
        page_ptr, num_pages);

    for (i = 0; (i < NV_DEFAULT_LIST_PAGE_COUNT) && (i < num_pages); i++)
    {
        nv_printf(NV_DBG_MEMINFO,
            "NVRM: VM:       v 0x%08lx p 0x%08lx d 0x%08lx: count %d flags 0x%x\n", 
            page_ptr->virt_addr,
            page_ptr->phys_addr,
            page_ptr->dma_addr,
            (NV_GET_PAGE_STRUCT(page_ptr->phys_addr))->count,
            (NV_GET_PAGE_STRUCT(page_ptr->phys_addr))->flags);
        page_ptr++;
    }
}
#endif


/* cache flushing
 * when allocating new pages, we convert the kernel mapping from cached to
 * uncached to avoid cache aliasing. one problem with this is that cpus can 
 * still contain data cached from these pages, in addition to stale ptes that
 * are cached and think the pages are still cached. normally, the cpu's self
 * snoop (SS) capability would catch this between cpus, but if the pages are
 * mapped through the agp aperture, SS is not capable of detecting these
 * conflicts. we try to rely on the kernel's provided interfaces when possible,
 * but need additional flushing on earlier kernels.
 */

/* wrap CACHE_FLUSH so we can pass it to smp_call_function */
static void cache_flush(void *p)
{
    CACHE_FLUSH();
}

/*
 * 2.4 kernels handle flushing in the change_page_attr() call, but kernels 
 * earlier than 2.4.27 don't flush on cpus that support Self Snoop, so we
 * manually flush on these kernels (actually, we go ahead and flush on all
 * 2.4 kernels, as it's possible some others may be missing this fix and
 * we'd prefer to be a little slower flushing caches than hanging the 
 * system. 2.6 kernels split the flushing out to a seperate call, 
 * global_flush_tlb(), so we rely on that.
 */
static void nv_flush_caches(void)
{
#if defined(KERNEL_2_4)
#ifdef CONFIG_SMP
    smp_call_function(cache_flush, NULL, 1, 1);
#endif
    cache_flush(NULL);
#else
    global_flush_tlb();
#endif
}

/*
 * DMA buffer management
 * nv_vm_malloc is used to allocate any memory used for dma purposes.
 * for various historical reasons, unless otherwise requested, we allocate
 * memory a page at a time.
 *
 * some cases where allocations have caused system problems:
 *
 *   allocating buffers with vmalloc can exhaust the kernel virtual address
 *     space. in some cases, the kernel's virtual address space can shrink
 *     down to < 128 megs on i386 systems.
 *     additionally, newer kernel api documents specifically forbid using
 *     vmalloc'd memory for dma.
 *
 *   allocating too many dma buffers via get_free_pages with order greater
 *     than 0 has also been shown to cause problems. too many large such
 *     allocations fragment the free_page pools, such that only single pages
 *     are left. at this point, creating new processes fails, as the task
 *     struct is 2 physical pages, which cannot be allocated. (this is not
 *     true with newer kernels, but is necessary for backwards compatibility)
 *
 * nv-linux.h defines which GFP_ flags to send to get_free_pages, based on
 * the processor platform. this is to insure we get 32-bit addresses when
 * possible.
 *
 * additionally, some platforms require us to allocate 64-bit addresses,
 * then remap them through an iommu/swiotlb to get 32-bit addresses. this
 * is handled here as well.
 *
 * the official kernel interface for allocating dma memory is to use
 * pci_alloc_consistent. this will abstract many of the 32-bit issues from us
 * mentioned above and will allow us to clean our code up quite a bit. we're
 * migrating towards this interface, but it's not quite stable in all cases
 * yet. em64t is one such current case.
 *
 * note that there's a subtle kernel interaction with regards to bookkeeping
 * on these pages. So long as the pages are marked reserved, the kernel won't
 * touch them (alter the usage count on them). this leads to a subtle problem
 * with mmap. Normally, allocating the pages would set the count to 1, then 
 * mmaping them would bump the count up to 2. The order of unmapping and freeing
 * the pages wouldn't matter, as they wouldn't really be considered free by the
 * kernel until the count dropped back to 0. Since the kernel won't touch the
 * count when the page is reserved, we need to be careful about this order and
 * unreserving the pages. if we unreserve the pages while freeing them, and the
 * munmap comes later, the munmap code path will attempt a second free on the 
 * same pages. We also don't have a lot of control over which comes first, 
 * sometimes we'll get called to free the pages first, sometimes we'll get 
 * called to munmap them first. Oh, and we'll get vma open/close calls every 
 * time the process is cloned, then execv'd, and munmap == vma close.
 * so, we keep our own count of the allocation usage, and don't unreserve the
 * pages until our count drops to 0. this should currently happen in either
 * vma_release or nv_vm_free, both of which will be followed by a kernel attempt
 * to free the page. Since the page will finally be unreserved, the kernel will
 * reduce the count to 0 and successfully free the page for us, only once.
 */

int nv_vm_malloc_pages(
    nv_alloc_t       *at
)
{
    /* point page_ptr at the start of the actual page list */
    nv_pte_t *page_ptr = at->page_table;
    int i;
    unsigned long virt_addr = 0, phys_addr;

    nv_printf(NV_DBG_MEMINFO, "NVRM: VM: nv_vm_malloc_pages: %d pages\n",
        at->num_pages);

    // allocate and prep contiguous pages up front if necessary
    if (NV_ALLOC_MAPPING_CONTIG(at->flags))
    {
#if defined(NV_SWIOTLB)
        if (nv_swiotlb && at->num_pages > 64)
        {
            nv_printf(NV_DBG_SETUP,
                "NVRM: VM: swiotlb can't support large contiguous allocation: %d pages\n",
                at->num_pages);
            return -1;
        }
#endif
        at->order = nv_calc_order(at->num_pages * PAGE_SIZE);
        nv_printf(NV_DBG_MEMINFO, "NVRM: VM:    contiguous, order %d\n", at->order);
        NV_GET_FREE_PAGES(virt_addr, at->order); 
        if (virt_addr == 0)
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: VM: nv_vm_malloc_pages: failed to allocate contiguous memory\n");
            return -1;
        }

#if defined(NV_SG_MAP_BUFFERS)
        // for amd 64-bit platforms, remap pages to make them 32-bit addressable
        // in this case, we need the final remapping to be contiguous, so we
        // have to do the whole mapping at once, instead of page by page
        if (nv_sg_map_buffer(at->dev, &at->sg_list[0],
                             (void *) virt_addr, at->num_pages))
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: VM: nv_vm_malloc_pages: failed to sg map contiguous pages\n");
            NV_FREE_PAGES(virt_addr, at->order);
            return -1;
        }
#endif
    }
    else if (NV_ALLOC_MAPPING_VMALLOC(at->flags))
    {
        void *virt_kptr = NULL;
        at->size = at->num_pages * PAGE_SIZE;
        nv_printf(NV_DBG_MEMINFO, "NVRM: VM:    vmalloc, size 0x%x\n", at->size);
        NV_VMALLOC(virt_kptr, at->size);
        virt_addr = (unsigned long) virt_kptr;
        if (virt_addr == 0)
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: nv_vm_malloc failed to allocate vmalloc memory\n");
            return -1;
        }
    }

    for (i = 0; i < at->num_pages; i++) 
    {
        if (!NV_ALLOC_MAPPING_CONTIG(at->flags) && !NV_ALLOC_MAPPING_VMALLOC(at->flags))
        { 
            NV_GET_FREE_PAGES(virt_addr, 0); 
            if (virt_addr == 0)
            {
                nv_printf(NV_DBG_ERRORS,
                    "NVRM: VM: nv_vm_malloc_pages: failed to allocate a page\n");
                goto failed;
            }
        }

        phys_addr = nv_get_kern_phys_address(virt_addr);
        if (phys_addr == 0)
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: nv_vm_malloc failed to lookup physical address\n");
            goto failed;
        }

        /* lock the page for dma purposes */
        SetPageReserved(NV_GET_PAGE_STRUCT(phys_addr));

        page_ptr = &at->page_table[i];
        page_ptr->phys_addr = phys_addr;
        page_ptr->virt_addr = virt_addr;
        page_ptr->dma_addr = page_ptr->phys_addr;

        if (!NV_ALLOC_MAPPING_CACHED(at->flags))
            NV_SET_PAGE_ATTRIB_UNCACHED(page_ptr);

#if defined(NV_SG_MAP_BUFFERS)
        if (!NV_ALLOC_MAPPING_CONTIG(at->flags))
        {
            if (nv_sg_map_buffer(at->dev, &at->sg_list[i], 
                                 __va(page_ptr->phys_addr), 1))
            {
                nv_printf(NV_DBG_ERRORS,
                    "NVRM: VM: nv_vm_malloc_pages: failed to sg map pages\n");
                goto failed;
            }
        }
        nv_sg_load(&at->sg_list[i], page_ptr);
#endif
        virt_addr += PAGE_SIZE;
    }

    nv_vm_list_page_count(at->page_table, at->num_pages);

    nv_flush_caches();

    return 0;

failed:

    for (; i >= 0; i--)
    {
        page_ptr = &at->page_table[i];

        // if we failed when allocating this page, skip over it
        // but if we failed pci_map_sg, make sure to free this page
        if (page_ptr && page_ptr->virt_addr)
        {
            NV_UNLOCK_PAGE(page_ptr);
#if defined(NV_SG_MAP_BUFFERS)
            if (!NV_ALLOC_MAPPING_CONTIG(at->flags))
                nv_sg_unmap_buffer(at->dev, &at->sg_list[i], page_ptr);
#endif
            if (!NV_ALLOC_MAPPING_CACHED(at->flags))
                NV_SET_PAGE_ATTRIB_CACHED(page_ptr);
            if (!NV_ALLOC_MAPPING_CONTIG(at->flags) && !NV_ALLOC_MAPPING_VMALLOC(at->flags))
                NV_FREE_PAGES(page_ptr->virt_addr, 0);
        }
    }
    nv_flush_caches();

    if (NV_ALLOC_MAPPING_CONTIG(at->flags))
    {
        page_ptr = at->page_table;
#if defined(NV_SG_MAP_BUFFERS)
        nv_sg_unmap_buffer(at->dev, &at->sg_list[0], page_ptr);
#endif
        NV_FREE_PAGES(page_ptr->virt_addr, at->order);
    }
    else if (NV_ALLOC_MAPPING_VMALLOC(at->flags))
    {
        page_ptr = at->page_table;
        NV_VFREE((void *) page_ptr->virt_addr, at->size);
    }

    return -1;
}

// unlock the pages we've locked down for dma purposes
void nv_vm_unlock_pages(
    nv_alloc_t       *at
)
{
    nv_pte_t *page_ptr;
    int i;

    nv_printf(NV_DBG_MEMINFO, "NVRM: VM: nv_vm_unlock_pages: %d pages\n",
        at->num_pages);

    if (at->page_table == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: VM: nv_vm_unlock_pages: NULL page table!");
        return;
    }

    nv_vm_list_page_count(at->page_table, at->num_pages);

    for (i = 0; i < at->num_pages; i++)
    {
        page_ptr = &at->page_table[i];
        NV_UNLOCK_PAGE(page_ptr);
    }
}

void nv_vm_free_pages(
    nv_alloc_t       *at
)
{
    nv_pte_t *page_ptr;
    int i;

    nv_printf(NV_DBG_MEMINFO, "NVRM: VM: nv_vm_free_pages: %d pages\n",
        at->num_pages);

    if (at->page_table == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: VM: nv_vm_free_pages: NULL page table!");
        return;
    }
    nv_vm_list_page_count(at->page_table, at->num_pages);

    for (i = 0; i < at->num_pages; i++)
    {
        page_ptr = &at->page_table[i];
#if defined(NV_SG_MAP_BUFFERS)
        if (!NV_ALLOC_MAPPING_CONTIG(at->flags))
            nv_sg_unmap_buffer(at->dev, &at->sg_list[i], page_ptr);
#endif
        if (!NV_ALLOC_MAPPING_CACHED(at->flags))
            NV_SET_PAGE_ATTRIB_CACHED(page_ptr);
        if (!NV_ALLOC_MAPPING_CONTIG(at->flags) && !NV_ALLOC_MAPPING_VMALLOC(at->flags))
            NV_FREE_PAGES(page_ptr->virt_addr, 0);
    }
    nv_flush_caches();

    if (NV_ALLOC_MAPPING_CONTIG(at->flags))
    {
        page_ptr = at->page_table;
#if defined(NV_SG_MAP_BUFFERS)
        nv_sg_unmap_buffer(at->dev, &at->sg_list[0], page_ptr);
#endif
        NV_FREE_PAGES(page_ptr->virt_addr, at->order);
    }
    else if (NV_ALLOC_MAPPING_VMALLOC(at->flags))
    {
        page_ptr = at->page_table;
        NV_VFREE((void *) page_ptr->virt_addr, at->size);
    }
}

