/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */



/******************* Operating System AGP Routines *************************\
*                                                                           *
* Module: os-agp.c                                                          *
*   interface to the operating system's native agp support                  *
*                                                                           *
\***************************************************************************/

#define  __NO_VERSION__
#include "nv-misc.h"

#include "os-interface.h"
#include "nv-linux.h"

#ifdef AGPGART

typedef struct {
    agp_memory *ptr;
    int num_pages;
    int offset;
#if defined(NVCPU_IA64)
    struct vm_area_struct *vma;
    void  *handle;
#endif
} agp_priv_data;

typedef struct {
    void *aperture;
#ifdef CONFIG_MTRR
    int  mtrr;
#endif
    int  ready;
} agp_gart;

agp_kern_info         agpinfo;
agp_gart              gart;
const drm_agp_t       *drm_agp_p;

#if defined(CONFIG_MTRR)
#define MTRR_DEL(gart) if ((gart).mtrr > 0) mtrr_del((gart).mtrr, 0, 0);
#else
#define MTRR_DEL(gart)
#endif

#endif /* AGPGART */

BOOL KernInitAGP(
    nv_state_t *nv,
    void **ap_phys_base,
    void **ap_mapped_base,
    U032 *apsize
)
{
#ifndef AGPGART
    return 1;
#else
    U032  agp_rate;
    U032  agp_sba;
    U032  agp_fw;
    void *bitmap;
    U032 bitmap_size;

    memset( (void *) &gart, 0, sizeof(agp_gart));

    if (!(drm_agp_p = inter_module_get_request("drm_agp", "agpgart")))
        return 1;

    /* NOTE: from here down, return an error code of '-1'
     * that indicates that agpgart is loaded, but we failed to use it
     * in some way. This is so we don't try to use nvagp and lock up
     * the memory controller.
     */

    if (drm_agp_p->acquire())
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: backend in use\n");
        inter_module_put("drm_agp");
        return -1;
    }

    if (rm_read_registry_dword(nv, "NVreg", "ReqAGPRate", &agp_rate) == RM_ERROR)
        agp_rate = 0x00000007;
    agp_rate &= 0x00000007;

    if (rm_read_registry_dword(nv, "NVreg", "EnableAGPSBA", &agp_sba) == RM_ERROR)
        agp_sba = 1;
    agp_sba &= 0x00000001;

    if (rm_read_registry_dword(nv, "NVreg", "EnableAGPFW", &agp_fw) == RM_ERROR)
        agp_fw = 1;
    agp_fw &= 0x00000001;

#if defined(KERNEL_2_4)
    /*
     * The original Linux 2.4 AGP GART driver interface declared copy_info to
     * return nothing. This changed in Linux 2.5, which reports unsupported
     * chipsets via this function. If this Linux 2.4 kernels behaves the same
     * way, we have no way to know.
     */
    drm_agp_p->copy_info(&agpinfo);
#else
    if (drm_agp_p->copy_info(&agpinfo)) {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: AGPGART: kernel reports chipset as unsupported\n");
        goto failed;
    }
#endif

#ifdef CONFIG_MTRR
    /*
     * Failure to set a write-combining range on the AGP aperture may
     * be due to the presence of other memory ranges with conflicting
     * caching  attributes. Play safe and fail AGP initialization.
     */
    if ((gart.mtrr = mtrr_add(agpinfo.aper_base,
                              agpinfo.aper_size * 0x100000,
                              MTRR_TYPE_WRCOMB, 0x0)) < 0)
#endif
    {
        nv_printf(NV_DBG_ERRORS, 
            "NVRM: AGPGART: unable to set MTRR write-combining\n");
        goto failed;
    }

    /*
     * Just map a single page for drivers to osGetAddressInfo() will work.
     * This will allow a client to create a DMA context for the entire
     * aperture without mapping the entire aperture.
     */

    NV_IOREMAP(gart.aperture, agpinfo.aper_base, RM_PAGE_SIZE);
    if (!gart.aperture)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: unable to remap aperture\n");
        goto failed;
    }

    // allocate and set the bitmap for tracking agp allocations
    bitmap_size = (agpinfo.aper_size * 0x100000)/PAGE_SIZE/8;
    if (os_alloc_mem(&bitmap, bitmap_size))
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: unable to allocate bitmap\n");
        NV_IOUNMAP(gart.aperture, RM_PAGE_SIZE);
        goto failed;
    }

    os_mem_set(bitmap, 0xff, bitmap_size);
    if (rm_set_agp_bitmap(nv, bitmap))
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: unable to set bitmap\n");
        os_free_mem(bitmap);
        NV_IOUNMAP(gart.aperture, RM_PAGE_SIZE);
        goto failed;
    }

    if (!agp_sba) agpinfo.mode &= ~0x00000200;
    if (!agp_fw)  agpinfo.mode &= ~0x00000010;

    if (!(agp_rate & 0x00000004)) agpinfo.mode &= ~0x00000004;
    if (!(agp_rate & 0x00000002)) agpinfo.mode &= ~0x00000002;
    
    drm_agp_p->enable(agpinfo.mode);

    *ap_phys_base   = (void*) agpinfo.aper_base;
    *ap_mapped_base = (void*) gart.aperture;
    *apsize         = (agpinfo.aper_size * 0x100000) - 1;

    gart.ready = 1;

    return 0;

failed:
    MTRR_DEL(gart); /* checks gart.mtrr */
    drm_agp_p->release();
    inter_module_put("drm_agp");

    return -1;

#endif /* AGPGART */
}

BOOL KernTeardownAGP(
    nv_state_t *nv
)
{
#ifndef AGPGART
    return 1;
#else
    void *bitmap;

    /* sanity check to make sure we should actually be here. */
    if (!gart.ready)
    {
        nv_printf(NV_DBG_ERRORS, 
            "NVRM: AGPGART: not ready so not releasing backend\n");
        return 0;
    }

    if (gart.aperture != NULL)
    {
        MTRR_DEL(gart);
        NV_IOUNMAP(gart.aperture, RM_PAGE_SIZE);
    }

    drm_agp_p->release();

    inter_module_put("drm_agp");

    if (rm_clear_agp_bitmap(nv, &bitmap))
    {
        nv_printf(NV_DBG_WARNINGS, "NVRM: AGPGART: failed to clear bitmap\n");
        return 1;
    }

    os_free_mem(bitmap);

    nv_printf(NV_DBG_SETUP, "NVRM: AGPGART: backend released\n");
    return 0;
#endif /* AGPGART */
}


RM_STATUS KernAllocAGPPages(
    nv_state_t *nv,
    void **pAddress,
    U032 PageCount,
    void **pPriv_data,
    U032 *Offset
)
{
#ifndef AGPGART
    *pAddress = (void*) 0;
    return RM_ERROR;
#else
    agp_memory *ptr;
    int err;
    agp_priv_data *data;
    RM_STATUS status;

    if (!gart.ready)
    {
        *pAddress = (void*) 0;
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: backend not ready\n");
        return RM_ERROR;
    }
    
    if (rm_alloc_agp_bitmap(nv, PageCount, Offset))
    {
        *pAddress = (void*) 0;
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: failed to allocate agp offset\n");
        return RM_ERROR;
    }

    ptr = drm_agp_p->allocate_memory(PageCount, AGP_NORMAL_MEMORY);
    if (ptr == NULL)
    {
        *pAddress = (void*) 0;
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: no pages available\n");
        return RM_ERR_NO_FREE_MEM;
    }
    
    err = drm_agp_p->bind_memory(ptr, *Offset);
    if (err)
    {
        // this happens a lot when the aperture itself fills up..
        // not a big deal, so don't alarm people with an error message
        nv_printf(NV_DBG_INFO, "NVRM: AGPGART: unable to bind %lu pages\n",
            (unsigned long) PageCount);
        goto fail;
    }

    /* return the agp aperture address */ 
    *pAddress = (void *) (agpinfo.aper_base + (*Offset << PAGE_SHIFT));

    status = os_alloc_mem((void **)&data, sizeof(agp_priv_data));
    if (status != RM_OK)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: memory allocation failed\n");
        drm_agp_p->unbind_memory(ptr);
        goto fail;
    }

    data->ptr = ptr;
    data->num_pages = PageCount;
    data->offset = *Offset;

    *pPriv_data = data;
    
    nv_printf(NV_DBG_INFO, "NVRM: AGPGART: allocated %lu pages\n",
           (unsigned long) PageCount);
    return RM_OK;

fail:
    drm_agp_p->free_memory(ptr);
    *pAddress = (void*) 0;

    return RM_ERROR;
#endif /* AGPGART */
}

#if !defined(NVCPU_IA64) || (LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 9))

RM_STATUS 
KernMapAGPPages(
    void *pvma,
    void *priv_data
)
{
#ifndef AGPGART
    return RM_ERROR;
#else
    struct vm_area_struct *vma = (struct vm_area_struct *) pvma;
    agp_priv_data *agp_data = (agp_priv_data *) priv_data;
    unsigned long agp_addr;
    int err;

    if ((agp_data == NULL) || (vma == NULL))
        return RM_ERROR;

    agp_addr = agpinfo.aper_base + (agp_data->offset << PAGE_SHIFT);

#if defined(NVCPU_IA64)
    err = NV_REMAP_PAGE_RANGE(vma->vm_start, (size_t) agp_addr, 
                       agp_data->num_pages << PAGE_SHIFT, vma->vm_page_prot);
#else
    err = NV_REMAP_PAGE_RANGE(vma->vm_start, (size_t) agp_addr, 
                       agp_data->num_pages << PAGE_SHIFT, PAGE_SHARED);
#endif
        
    if (err)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: AGPGART: unable to remap %lu pages\n",
            (unsigned long)agp_data->num_pages);
        drm_agp_p->unbind_memory(agp_data->ptr);
        goto fail;
    }
    
#if defined(NVCPU_IA64)
    vma->vm_pgoff = agp_addr >> PAGE_SHIFT;
    agp_data->vma = vma;
    agp_register_map(vma);
#endif

    return RM_OK;

fail:
    /* XXX what else needs to happen here? */
    return RM_ERROR;
#endif /* AGPGART */
}
#endif

RM_STATUS 
KernMapAGPNopage(
    void *address,
    void *pvma,
    void *priv_data,
    void **ppage)
{
#ifndef AGPGART
    return RM_OK;
#else
    struct vm_area_struct *vma = (struct vm_area_struct *) pvma;
    agp_priv_data *agp_data = (agp_priv_data *) priv_data;
    agp_memory *agp_memory_ptr = agp_data->ptr;
    struct page *page_ptr;
    unsigned long phys_addr;
    int idx;

    idx = ((unsigned long)address - vma->vm_start) >> PAGE_SHIFT;
    /* There was an interface change durring the 2.4.18 series.  Earlier
      * 2.4.18 ports patches used the earlier interface but most shipping
      * kernels use the new interface.
      */
#if defined(NVCPU_IA64) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 18))
    phys_addr = agp_memory_ptr->memory[idx];
#else
    phys_addr = (agp_memory_ptr->memory[idx] & 0xffffff) << 12;
#endif
    page_ptr = virt_to_page(__va(phys_addr));
    get_page(page_ptr);
    *ppage = page_ptr;

#if defined(NVCPU_IA64)
    {
        /*
         * Reliability is dramatically improved when invalidating the cache 
         * lines prior to using write combined mappings.  The "fc" instruction
         * is specified to affect at least 32 bytes, thus the 32 byte increment
         * below.
         */
        int i;
        char *cp = (char *)__va(phys_addr);
        for (i = 0; i < PAGE_SIZE/32; i++)
        {
            ia64_fc (cp);
            cp += 32;
        }
        ia64_sync_i(); /* paranoia */
    }
#endif

    return RM_OK;
#endif
}


RM_STATUS KernFreeAGPPages(
    nv_state_t *nv,
    void **pAddress,
    void *priv_data
)
{
#ifndef AGPGART
    return RM_OK;
#else
    agp_memory *ptr;
    agp_priv_data *data = priv_data;
    
#if defined(NVCPU_IA64) && (LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 9))
    agp_unregister_map(data->vma);
#endif

    rm_free_agp_bitmap(nv, data->num_pages, data->offset);

    ptr = data->ptr;
    os_free_mem(data);

    if (!(ptr == NULL))
    {
        size_t pages = ptr->page_count;

        drm_agp_p->unbind_memory(ptr);
        drm_agp_p->free_memory(ptr);

        nv_printf(NV_DBG_INFO, "NVRM: AGPGART: freed %ld pages\n",
            (unsigned long)pages);
        return RM_OK;
    }
    return RM_ERROR;
#endif
}


